@@ -1,11 +1,12 @@
 /*
  *	file.c
  *
- *	File Access Module
+ *	file accessing module
  */
 
 #include <fcntl.h>
 #include "fd.h"
+#include "termio.h"
 #include "func.h"
 #include "kanji.h"
 
@@ -32,9 +33,6 @@ extern int checkdrive __P_((int));
 #endif
 
 extern int mark;
-extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
 extern int physical_path;
@@ -110,6 +108,7 @@ char *path, *file;
 	path[0] = '.';
 	path[1] = _SC_;
 	strcpy(&(path[2]), file);
+
 	return(path);
 }
 #endif	/* _USEDOSEMU */
@@ -135,6 +134,7 @@ gid_t gid;
 	else if (isgroupmember(gid)) mode >>= 3;
 #endif	/* !NOUID */
 	if (dir && !(mode & F_ISEXE)) mode &= ~(F_ISRED | F_ISWRI);
+
 	return(mode & 007);
 }
 
@@ -150,14 +150,15 @@ namelist *namep;
 	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
 	namep -> flags = 0;
-	if ((st.st_mode & S_IFMT) == S_IFDIR) namep -> flags |= F_ISDIR;
-	if ((lst.st_mode & S_IFMT) == S_IFLNK) namep -> flags |= F_ISLNK;
+	if (s_isdir(&st)) namep -> flags |= F_ISDIR;
+	if (s_islnk(&lst)) namep -> flags |= F_ISLNK;
 
 	if (isdisplnk(dispmode))
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #if	!MSDOS
 	if ((lst.st_mode & S_IFMT) == S_IFCHR
-	|| (lst.st_mode & S_IFMT) == S_IFBLK) namep -> flags |= F_ISDEV;
+	|| (lst.st_mode & S_IFMT) == S_IFBLK)
+		namep -> flags |= F_ISDEV;
 #endif
 
 	namep -> st_mode = lst.st_mode;
@@ -181,12 +182,8 @@ namelist *namep;
 #else
 		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
 #endif
-
-	if (isfile(namep) && namep -> st_size) {
-		totalsize += getblock(namep -> st_size);
-		if (ismark(namep)) marksize += getblock(namep -> st_size);
-	}
 	namep -> tmpflags |= F_STAT;
+
 	return(0);
 }
 
@@ -213,9 +210,11 @@ CONST VOID_P vp2;
 	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '\0') return(-1);
 	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '\0') return(1);
 	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '.'
-	&& namep1 -> name[2] == '\0') return(-1);
+	&& namep1 -> name[2] == '\0')
+		return(-1);
 	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '.'
-	&& namep2 -> name[2] == '\0') return(1);
+	&& namep2 -> name[2] == '\0')
+		return(1);
 
 	switch (sorton & 7) {
 		case 5:
@@ -307,8 +306,8 @@ CONST VOID_P vp2;
 			tmp = 0;
 			break;
 	}
-
 	if (sorton > 7) tmp = -tmp;
+
 	return(tmp);
 }
 #endif
@@ -331,8 +330,7 @@ char *arcregstr;
 	}
 #ifndef	_NOARCHIVE
 	else if (arcregstr) while ((dp = Xreaddir(dirp))) {
-		if (stat2(dp -> d_name, &st) < 0
-		|| (st.st_mode & S_IFMT) == S_IFDIR) continue;
+		if (stat2(dp -> d_name, &st) < 0 || s_isdir(&st)) continue;
 
 		tmp.name = dp -> d_name;
 		tmp.flags = 0;
@@ -343,6 +341,7 @@ char *arcregstr;
 	}
 #endif
 	else dp = Xreaddir(dirp);
+
 	return(dp);
 }
 
@@ -362,11 +361,10 @@ char *buf;
 	}
 	if (!homedir) {
 		if (!(cp = gethomedir())) return(-1);
-		if (
 #ifndef	_NODOSDRIVE
-		dospath2(cp) ||
+		if (dospath2(cp)) return(-1);
 #endif
-		_chdir2(cp) < 0) return(-1);
+		if (_chdir2(cp) < 0) return(-1);
 		homedir = getwd2();
 		if (_chdir2(cwd) < 0) lostcwd(cwd);
 	}
@@ -385,6 +383,7 @@ char *buf;
 	free(homedir);
 	homedir = NULL;
 #endif
+
 	return((cp) ? 1 : 0);
 }
 
@@ -416,13 +415,13 @@ char *dir;
 			}
 			else if (stat2(dir, &st) < 0) return(-1);
 		}
-		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+		if (!s_isdir(&st)) {
 			errno = ENOTDIR;
 			return(-1);
 		}
 	}
-
 	entryhist(1, realpath2(dir, tmp, 0), 1);
+
 	return(0);
 }
 
@@ -434,7 +433,7 @@ struct stat *stp;
 	int ret, duperrno;
 
 	if (Xlstat(path, &st) < 0) return(-1);
-	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
+	if (s_islnk(&st)) return(1);
 
 	ret = 0;
 #ifdef	FAKEUNINIT
@@ -467,7 +466,7 @@ struct stat *stp;
 #endif
 	}
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != 0xffffffff) {
+	if (stp -> st_flags != (u_long)-1) {
 # ifndef	_NODOSDRIVE
 		if (dospath(path, NULL)) {
 			duperrno = EACCES;
@@ -481,7 +480,7 @@ struct stat *stp;
 		}
 	}
 #endif
-	dummy.st_mode = 0xffff;
+	dummy.st_mode = (u_short)-1;
 	if (stp -> st_mode != dummy.st_mode) {
 		stp -> st_mode &= ~S_IFMT;
 		stp -> st_mode |= (st.st_mode & S_IFMT);
@@ -504,6 +503,7 @@ struct stat *stp;
 	}
 #endif	/* !NOUID */
 	if (ret < 0) errno = duperrno;
+
 	return(ret);
 }
 
@@ -517,6 +517,7 @@ char *path;
 	dosdrive = 0;
 	n = Xchdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -530,6 +531,7 @@ char *path;
 	dosdrive = 0;
 	cp = Xgetwd(path);
 	dosdrive = dupdosdrive;
+
 	return(cp);
 }
 
@@ -543,6 +545,7 @@ struct stat *stp;
 	dosdrive = 0;
 	n = Xlstat(path, stp);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -556,6 +559,7 @@ int mode;
 	dosdrive = 0;
 	n = Xmkdir(path, mode);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -568,6 +572,7 @@ char *path;
 	dosdrive = 0;
 	n = Xrmdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 #endif	/* !_NODOSDRIVE */
@@ -596,24 +601,39 @@ char *path;
 	errno = duperrno;
 }
 
-int safewrite(fd, buf, size)
-int fd;
-char *buf;
-int size;
+#ifndef	NODIRLOOP
+int issamebody(src, dest)
+char *src, *dest;
 {
-	int n;
+	struct stat st1, st2;
 
-	n = Xwrite(fd, buf, size);
-#if	MSDOS
-	if (n >= 0 && n < size) {
-		n = -1;
-		errno = ENOSPC;
-	}
-#else
-	if (n < 0 && errno == EINTR) n = size;
-#endif
-	return(n);
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
 }
+#endif	/* !NODIRLOOP */
+
+#ifndef	NOSYMLINK
+int cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+# ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+# endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+
+	path[len] = '\0';
+
+	return(Xsymlink(path, dest));
+}
+#endif	/* !NOSYMLINK */
 
 /*ARGSUSED*/
 static int NEAR cpfile(src, dest, stp1, stp2)
@@ -624,43 +644,54 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, duperrno;
+	int i, fd1, fd2, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
 	if (stp2 && !(stp2 -> st_mode & S_IWRITE))
 		Xchmod(src, stp2 -> st_mode | S_IWRITE);
 #endif
-	if ((stp1 -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, buf, BUFSIZ - 1)) < 0) return(-1);
-		buf[i] = '\0';
-		return(Xsymlink(buf, dest));
-	}
-	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
-		return(-1);
-#if	MSDOS
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
-		stp1 -> st_mode | S_IWRITE);
-#else
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
+#ifndef	NOSYMLINK
+	if (s_islnk(stp1)) return(cpsymlink(src, dest));
 #endif
-	if (fd2 < 0) {
+
+	flags = (O_BINARY | O_RDONLY);
+	mode = stp1 -> st_mode;
+	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
+
+	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
+#if	MSDOS
+	mode |= S_IWRITE;
+#endif
+#ifndef	NODIRLOOP
+	if (issamebody(src, dest)) {
+		flags |= O_EXCL;
+		if (Xunlink(dest) < 0) {
+			Xclose(fd1);
+			return(-1);
+		}
+	}
+#endif
+	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
 		Xclose(fd1);
 		return(-1);
 	}
 
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 	for (;;) {
-		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
-		duperrno = errno;
-		if (i < BUFSIZ) break;
-		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
+		if ((i = sureread(fd1, buf, BUFSIZ)) <= 0) {
 			duperrno = errno;
 			break;
 		}
+		if (surewrite(fd2, buf, i) < 0) {
+			i = -1;
+			duperrno = errno;
+			break;
+		}
+		if (i < BUFSIZ) break;
 	}
-	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) duperrno = errno;
 
 	Xclose(fd2);
 	Xclose(fd1);
@@ -671,7 +702,7 @@ struct stat *stp1, *stp2;
 		return(-1);
 	}
 #ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
+	stp1 -> st_flags = (u_long)-1;
 #endif
 #ifdef	_USEDOSCOPY
 	if (touchfile(dest, stp1) < 0) return(-1);
@@ -680,6 +711,7 @@ struct stat *stp1, *stp2;
 	if (inheritcopy && touchfile(dest, stp1) < 0) return(-1);
 # endif
 #endif
+
 	return(0);
 }
 
@@ -723,13 +755,13 @@ char *src, *dest;
 struct stat *stp1, *stp2;
 {
 	if (Xrename(src, dest) >= 0) return(0);
-	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
-		return(-1);
+	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
 #ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
+	stp1 -> st_flags = (u_long)-1;
 #endif
+
 	return (touchfile(dest, stp1));
 }
 
@@ -771,7 +803,7 @@ int len;
 	int i, j, c;
 
 	if (!buf) {
-		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
+		for (i = 0; i < (int)sizeof(seq) / sizeof(char); i++) {
 			j = genrand(sizeof(seq) / sizeof(char));
 			c = seq[i];
 			seq[i] = seq[j];
@@ -785,6 +817,7 @@ int len;
 		}
 		buf[i] = '\0';
 	}
+
 	return(buf);
 }
 
@@ -823,9 +856,9 @@ char *dir;
 		}
 	}
 	if (!tmpfilename) {
-		n = sizeof(TMPPREFIX) - 1;
+		n = (int)sizeof(TMPPREFIX) - 1;
 		strncpy(cp, TMPPREFIX, n);
-		len = sizeof(path) - 1 - (cp - path);
+		len = (int)sizeof(path) - 1 - (cp - path);
 		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 		len -= n;
 		genrandname(NULL, 0);
@@ -844,7 +877,7 @@ char *dir;
 	}
 
 	strncpy((cp = strcatdelim(path)), dir, n);
-	len = sizeof(path) - 1 - (cp - path);
+	len = (int)sizeof(path) - 1 - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	len -= n;
 	genrandname(NULL, 0);
@@ -864,6 +897,7 @@ char *dir;
 		nodosrmdir(path);
 		errno = n;
 	}
+
 	return(-1);
 }
 
@@ -884,6 +918,7 @@ char *dir;
 	}
 	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
+
 	return(0);
 }
 
@@ -896,7 +931,7 @@ char *file;
 	path[0] = '\0';
 	if (mktmpdir(path) < 0) return(-1);
 	cp = strcatdelim(path);
-	len = sizeof(path) - 1 - (cp - path);
+	len = (int)sizeof(path) - 1 - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	genrandname(NULL, 0);
 
@@ -913,6 +948,7 @@ char *file;
 	duperrno = errno;
 	rmtmpdir(NULL);
 	errno = duperrno;
+
 	return(-1);
 }
 
@@ -921,6 +957,7 @@ char *file;
 {
 	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
 		return(-1);
+
 	return(0);
 }
 
@@ -934,6 +971,7 @@ static int dormdir(path)
 char *path;
 {
 	if (isdotdir(path)) return(0);
+
 	return(Xrmdir(path));
 }
 
@@ -996,6 +1034,7 @@ char *dir, *file;
 		_exit(1);
 	}
 #endif
+
 	return(0);
 }
 
@@ -1009,6 +1048,7 @@ int drive;
 	path[1] = drive;
 	path[2] = '\0';
 	if (mktmpdir(path) < 0) return(NULL);
+
 	return(strdup2(path));
 }
 
@@ -1063,6 +1103,7 @@ int single;
 		return(-1);
 	}
 	*dirp = tmpdir;
+
 	return(drive);
 }
 
@@ -1097,6 +1138,7 @@ char *file;
 	struct stat st;
 
 	if (Xlstat(file, &st) < 0 && errno == ENOENT) return(0);
+
 	return(1);
 }
 
@@ -1134,6 +1176,7 @@ int dos, boundary, dirsize, ofs;
 		len += ofs;
 		if (lfn) return((len / LFNENTSIZ + 1) * DOSDIRENT + DOSDIRENT);
 	}
+
 	return(DOSDIRENT);
 }
 
@@ -1149,6 +1192,7 @@ int size, dos, boundary, dirsize, ofs;
 		size -= DOSDIRENT;
 		return((size / DOSDIRENT) * LFNENTSIZ - 1);
 	}
+
 	return(DOSBODYLEN);
 }
 
@@ -1164,6 +1208,7 @@ char *from, *to;
 	from = nodospath(fpath, from);
 	to = nodospath(tpath, to);
 #endif
+
 	return(Xrename(from, to));
 }
 
@@ -1210,8 +1255,9 @@ char *tmpdir, *old;
 					else break;
 				}
 				else {
-					if ((fd = Xopen(fname,
-					O_CREAT | O_EXCL, 0600))) {
+					fd = Xopen(fname,
+						O_CREAT | O_EXCL, 0600);
+					if (fd >= 0) {
 						Xclose(fd);
 						return(fname);
 					}
@@ -1225,6 +1271,7 @@ char *tmpdir, *old;
 		}
 	}
 	free(fname);
+
 	return(NULL);
 }
 
@@ -1235,6 +1282,7 @@ char *dir;
 	struct stat st;
 
 	if (Xlstat(dir, &st) < 0) return(getblocksize(dir));
+
 	return((off_t)st.st_size);
 }
 
@@ -1244,7 +1292,7 @@ off_t bsiz;
 {
 	struct stat st;
 	u_char ch, *tmp;
-	int i, c, n, fd;
+	int i, n, fd;
 
 	if (Xlstat(dir, &st) < 0
 	|| (fd = Xopen(dir, O_BINARY | O_RDONLY, 0666)) < 0)
@@ -1257,9 +1305,7 @@ off_t bsiz;
 			free(tmp);
 			return(NULL);
 		}
-		while ((c = Xread(fd, (char *)&ch, sizeof(ch))) < 0
-		&& errno == EINTR);
-		if (c < 0) {
+		if (sureread(fd, &ch, sizeof(ch)) < 0) {
 			free(tmp);
 			return(NULL);
 		}
@@ -1312,21 +1358,21 @@ int fs;
 
 	switch (fs) {
 #if	!MSDOS
-		case 2:	/* IRIX File System */
+		case FSID_EFS:		/* IRIX File System */
 			dos = 0;
 			headbyte = 4;
 			boundary = 2;
 			dirsize = sizeof(u_long);
 			namofs = 0;
 			break;
-		case 3:	/* SystemV R3 File System */
+		case FSID_SYSV:		/* SystemV R3 File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 8;
 			dirsize = sizeof(u_short);
 			namofs = 0;
 			break;
-		case 6:	/* Linux File System */
+		case FSID_LINUX:	/* Linux File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 4;
@@ -1334,7 +1380,7 @@ int fs;
 			namofs = 3;
 			break;
 # ifndef	_NODOSDRIVE
-		case 7:	/* Windows95 File System on DOSDRIVE */
+		case FSID_DOSDRIVE:	/* Windows95 File System on DOSDRIVE */
 			dos = 3;
 			headbyte = -1;
 			boundary = LFNENTSIZ;
@@ -1343,7 +1389,7 @@ int fs;
 			break;
 # endif
 #endif	/* !MSDOS */
-		case 4:	/* MS-DOS File System */
+		case FSID_FAT:		/* MS-DOS File System */
 			dos = 1;
 #if	!MSDOS
 			headbyte = -1;
@@ -1352,7 +1398,7 @@ int fs;
 			dirsize = DOSDIRENT;
 			namofs = 0;
 			break;
-		case 5:	/* Windows95 File System */
+		case FSID_LFN:		/* Windows95 File System */
 			dos = 2;
 #if	!MSDOS
 			headbyte = -1;
@@ -1456,7 +1502,7 @@ int fs;
 	}
 
 #if	!MSDOS
-	if (fs != 2) entnum = NULL;	/* except IRIX File System */
+	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
 	else if (!(entnum = getentnum(".", persec))) {
 		warning(-1, ".");
 		restorefile(tmpdir, path, fnamp);
@@ -1482,13 +1528,13 @@ int fs;
 		size = realdirsiz(fnamelist[i],
 			dos, boundary, dirsize, namofs);
 		switch (fs) {
-			case 2:	/* IRIX File System */
+			case FSID_EFS:	/* IRIX File System */
 				if (totalptr > ptr + 1) ent -= totalptr;
 				else ent -= ptr + 1;
 				break;
-			case 3:	/* SystemV R3 File System */
-			case 4:	/* MS-DOS File System */
-			case 5:	/* Windows95 File System */
+			case FSID_SYSV:	/* SystemV R3 File System */
+			case FSID_FAT:	/* MS-DOS File System */
+			case FSID_LFN:	/* Windows95 File System */
 				ent = size;
 				break;
 			default:
