@@ -1,7 +1,7 @@
 /*
  *	pathname.c
  *
- *	Path Name Management Module
+ *	path name management module
  */
 
 #ifdef	FD
@@ -96,6 +96,7 @@ extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
+extern int stat2 __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 # if	MSDOS && !defined (_NOUSELFN)
 extern char *shortname __P_((char *, char *));
@@ -121,12 +122,23 @@ struct dirent *Xreaddir __P_((DIR *));
 static u_int NEAR getdosmode __P_((u_int));
 static time_t NEAR getdostime __P_((u_int, u_int));
 int Xstat __P_((char *, struct stat *));
+#  ifndef	NOSYMLINK
+#  define	Xlstat		Xstat
+#  endif
 # else	/* !MSDOS */
 # define	Xopendir	opendir
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
 # define	Xstat		stat
+#  ifndef	NOSYMLINK
+#  define	Xlstat		lstat
+#  endif
 # endif	/* !MSDOS */
+# ifdef	NOSYMLINK
+# define	stat2		Xstat
+# else
+int stat2 __P_((char *, struct stat *));
+# endif
 # ifdef	DJGPP
 char *Xgetwd __P_((char *));
 # else	/* !DJGPP */
@@ -141,19 +153,16 @@ char *Xgetwd __P_((char *));
 
 static char *NEAR getenvvar __P_((char *, int));
 static int NEAR setvar __P_((char *, char *, int));
-static int NEAR ismeta __P_((char *s, int, int, int));
+static int NEAR ismeta __P_((char *s, int, int, int, int));
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp __P_((char *, int));
 #else
 static int NEAR _regexp_exec __P_((char **, char *));
 #endif
-static char *NEAR catpath __P_((char *, char *, int *, int, int));
-#ifdef	NODIRLOOP
-static int NEAR _evalwild __P_((int, char ***, char *, char *, int));
-#else
-static int NEAR _evalwild __P_((int, char ***, char *, char *, int,
-		int, devino_t *));
-#endif
+static void NEAR addstrbuf __P_((strbuf_t *, char *, int));
+static void NEAR duplwild __P_((wild_t *, wild_t *));
+static void NEAR freewild __P_((wild_t *));
+static int NEAR _evalwild __P_((int, char ***, wild_t *));
 #ifndef	_NOUSEHASH
 static int NEAR calchash __P_((char *));
 static VOID NEAR inithash __P_((VOID_A));
@@ -237,6 +246,7 @@ char *s;
 	fputs(s, stderr);
 	fputs(": memory allocation error", stderr);
 	fputnl(stderr);
+
 	exit(2);
 }
 
@@ -251,6 +261,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -268,6 +279,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -280,6 +292,7 @@ ALLOC_T n, *sizep;
 		return(malloc2(*sizep));
 	}
 	while (n + 1 >= *sizep) *sizep *= 2;
+
 	return(realloc2(ptr, *sizep));
 }
 
@@ -293,6 +306,7 @@ char *s;
 	n = strlen(s);
 	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) error("malloc()");
 	memcpy(tmp, s, n + 1);
+
 	return(tmp);
 }
 
@@ -308,6 +322,7 @@ int n;
 	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) error("malloc()");
 	memcpy(tmp, s, i);
 	tmp[i] = '\0';
+
 	return(tmp);
 }
 
@@ -324,6 +339,7 @@ int c;
 		else if (isekana(s, i)) i++;
 #endif
 	}
+
 	return(NULL);
 }
 
@@ -334,6 +350,7 @@ char *s1, *s2;
 
 	for (i = 0; s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -345,6 +362,7 @@ int n;
 
 	for (i = 0; i < n && s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -386,6 +404,7 @@ char *dir;
 			return(NULL);
 		}
 	}
+
 	return(dirp);
 }
 
@@ -395,6 +414,7 @@ DIR *dirp;
 	free(dirp -> dd_buf);
 	free(dirp -> dd_path);
 	free(dirp);
+
 	return(0);
 }
 
@@ -468,10 +488,32 @@ struct stat *stp;
 	stp -> st_dev = stp -> st_ino = 0;
 	stp -> st_nlink = 1;
 	stp -> st_uid = stp -> st_gid = -1;
+
 	return(0);
 }
 # endif	/* MSDOS */
 
+# ifndef	NOSYMLINK
+int stat2(path, stp)
+char *path;
+struct stat *stp;
+{
+	int duperrno;
+
+	if (Xstat(path, stp) < 0) {
+		duperrno = errno;
+		if (Xlstat(path, stp) < 0
+		|| (stp -> st_mode & S_IFMT) != S_IFLNK) {
+			errno = duperrno;
+			return(-1);
+		}
+		stp -> st_mode &= ~S_IFMT;
+	}
+
+	return(0);
+}
+# endif	/* !NOSYMLINK */
+
 # ifdef	DJGPP
 char *Xgetwd(path)
 char *path;
@@ -481,6 +523,7 @@ char *path;
 
 	if (!(cp = (char *)getcwd(path, MAXPATHLEN))) return(NULL);
 	for (i = 0; cp[i]; i++) if (cp[i] == '/') cp[i] = _SC_;
+
 	return(cp);
 }
 # endif	/* !DJGPP */
@@ -500,6 +543,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(NULL);
 }
 
@@ -518,6 +562,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(cp);
 }
 #endif	/* MSDOS || (FD && !_NODOSDRIVE) */
@@ -534,9 +579,11 @@ char *s, *eol;
 		if (s[i] == _SC_) cp = &(s[i]);
 		if (iskanji1(s, i)) i++;
 	}
+
 	return(cp);
 #else
 	for (eol--; eol >= s; eol--) if (*eol == _SC_) return(eol);
+
 	return(NULL);
 #endif
 }
@@ -556,6 +603,7 @@ int ptr;
 
 	for (i = 0; s[i] && i < ptr; i++) if (iskanji1(s, i)) i++;
 	if (!s[i] || i > ptr) return(1);
+
 	return(!iskanji1(s, i));
 #else
 	return(1);
@@ -593,6 +641,7 @@ char *s;
 		*cp = _SC_;
 	}
 	*(++cp) = '\0';
+
 	return(cp);
 }
 
@@ -646,6 +695,7 @@ char *buf, *s1, *s2;
 		for (i = 0; s2[i] && i < len; i++) *(cp++) = s2[i];
 	}
 	*cp = '\0';
+
 	return(cp);
 }
 
@@ -665,6 +715,7 @@ char *s1, *s2;
 		s1++;
 		s2++;
 	}
+
 	return(0);
 }
 
@@ -686,6 +737,7 @@ int n;
 		s1++;
 		s2++;
 	}
+
 	return(0);
 }
 
@@ -694,7 +746,8 @@ int strpathcmp2(s1, s2)
 char *s1, *s2;
 {
 	if (pathignorecase) return(strcasecmp2(s1, s2));
-	else return(strcmp(s1, s2));
+
+	return(strcmp(s1, s2));
 }
 
 int strnpathcmp2(s1, s2, n)
@@ -702,7 +755,8 @@ char *s1, *s2;
 int n;
 {
 	if (pathignorecase) return(strncasecmp2(s1, s2, n));
-	else return(strncmp(s1, s2, n));
+
+	return(strncmp(s1, s2, n));
 }
 #endif	/* !PATHNOCASE */
 
@@ -736,6 +790,7 @@ int len;
 	cp = strndup2(ident, len);
 	env = getenv2(cp);
 	free(cp);
+
 	return(env);
 #endif
 }
@@ -759,6 +814,7 @@ int len;
 	ret = (putvarfunc) ? (*putvarfunc)(cp, len) : putenv(cp);
 #endif
 	if (ret < 0) free(cp);
+
 	return(ret);
 }
 
@@ -778,6 +834,7 @@ int isdotdir(s)
 char *s;
 {
 	if (s[0] == '.' && (!s[1] || (s[1] == '.' && !s[2]))) return(1);
+
 	return(0);
 }
 
@@ -788,6 +845,7 @@ char *s;
 	if (_dospath(s)) s += 2;
 #endif
 	if (*s != _SC_) return(NULL);
+
 	return(s);
 }
 
@@ -797,6 +855,7 @@ char *s;
 	char *cp;
 
 	if ((cp = strrdelim(s, 1))) return(cp + 1);
+
 	return(s);
 }
 
@@ -823,30 +882,41 @@ int *loginp, *restrictedp;
 	return(s);
 }
 
-static int NEAR ismeta(s, ptr, quote, len)
+static int NEAR ismeta(s, ptr, quote, len, flags)
 char *s;
-int ptr, quote, len;
+int ptr, quote, len, flags;
 {
 #ifdef	FAKEMETA
 	return(0);
 #else	/* !FAKEMETA */
-	if (s[ptr] != PMETA) return(0);
+	if (s[ptr] != PMETA || quote == '\'') return(0);
+
 	if (len >= 0) {
-		if (ptr + 1 >= len) return(0);
-		if (quote && s[ptr + 1] == quote && ptr + 2 >= len) return(0);
+		if (!(flags & EA_EOLMETA) && ptr + 1 >= len) return(0);
+# ifndef	BASHSTYLE
+	/* bash does not treat "\" as \ */
+		if (quote == '"' && s[ptr + 1] == quote && ptr + 2 >= len)
+			return(0);
+# endif
 	}
 	else {
-		if (!s[ptr + 1]) return(0);
-		if (quote) {
-			if (s[ptr + 1] == quote && !s[ptr + 2]) return(0);
-# ifdef	BSPATHDELIM
-			if (!strchr(DQ_METACHAR, s[ptr + 1])) return(0);
-# endif
-		}
-# ifdef	BSPATHDELIM
-		else if (!strchr(METACHAR, s[ptr + 1])) return(0);
+		if (!(flags & EA_EOLMETA) && !s[ptr + 1]) return(0);
+# ifndef	BASHSTYLE
+	/* bash does not treat "\" as \ */
+		if (quote == '"' && s[ptr + 1] == quote && !s[ptr + 2])
+			return(0);
 # endif
 	}
+
+# ifdef	BSPATHDELIM
+	if (quote == '"') {
+		if (!strchr(DQ_METACHAR, s[ptr + 1])) return(0);
+	}
+	else {	/* if (!quote || quote == '`') */
+		if (!strchr(METACHAR, s[ptr + 1])) return(0);
+	}
+# endif
+
 	return(1);
 #endif	/* !FAKEMETA */
 }
@@ -876,7 +946,7 @@ int len;
 			re[j++] = s[i];
 			continue;
 		}
-		else if (ismeta(s, i, '\0', len)) {
+		else if (ismeta(s, i, '\0', len, 0)) {
 			re[j++] = s[i++];
 			re[j++] = s[i];
 		}
@@ -905,6 +975,7 @@ int len;
 	}
 	re[j++] = '$';
 	re[j++] = '\0';
+
 	return(re);
 }
 #endif	/* _NOORIGGLOB */
@@ -927,6 +998,7 @@ int len;
 	s = cnvregexp(s, len);
 	re_comp(s);
 	free(s);
+
 	return((reg_t *)1);
 }
 
@@ -937,6 +1009,7 @@ char *s;
 int fname;
 {
 	if (fname && skipdotfile && *s == '.') return(0);
+
 	return(re_exec(s) > 0);
 }
 
@@ -968,6 +1041,7 @@ int len;
 		re = NULL;
 	}
 	free(s);
+
 	return(re);
 }
 
@@ -977,6 +1051,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(!regexec(re, s, 0, NULL, 0));
 }
 
@@ -1004,6 +1079,7 @@ int len;
 	s = cnvregexp(s, len);
 	re = regcmp(s, 0);
 	free(s);
+
 	return(re);
 }
 
@@ -1013,6 +1089,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(regex(re, s) ? 1 : 0);
 }
 
@@ -1032,7 +1109,7 @@ int len;
 	ALLOC_T size;
 	int i, j, n, pc, plen, metachar, quote;
 
-	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
+	skipdotfile = 0;
 	if (len < 0) len = strlen(s);
 	re = NULL;
 	paren = NULL;
@@ -1138,13 +1215,16 @@ int len;
 		else if (!quote && !metachar) switch (s[i]) {
 			case '?':
 				cp = wildsymbol1;
+				if (!n) skipdotfile++;
 				break;
 			case '*':
 				cp = wildsymbol2;
+				if (!n) skipdotfile++;
 				break;
 			case '[':
 				paren = c_realloc(NULL, 0, &size);
 				plen = 0;
+				if (!n) skipdotfile++;
 				break;
 			default:
 				break;
@@ -1185,6 +1265,7 @@ int len;
 		re = b_realloc(re, n, reg_t);
 		re[n] = NULL;
 	}
+
 	return((reg_t *)realloc2(re, (n + 1) * sizeof(reg_t)));
 }
 
@@ -1254,6 +1335,7 @@ char *s;
 	}
 	while (re[n1] == wildsymbol2) n1++;
 	if (re[n1] || s[n2]) return(0);
+
 	return(1);
 }
 
@@ -1263,6 +1345,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(_regexp_exec(re, s));
 }
 
@@ -1282,245 +1365,226 @@ reg_t *re;
 # endif		/* !USEREGCOMP */
 #endif		/* !USERE_COMP */
 
-static char *NEAR catpath(path, file, plenp, flen, isoverwrite)
-char *path, *file;
-int *plenp, flen, isoverwrite;
+static void NEAR addstrbuf(sp, s, len)
+strbuf_t *sp;
+char *s;
+int len;
 {
-	char *new;
-	int i, sc;
-
-#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-	if (*plenp >= 2 && _dospath(path)) i = 2;
-	else
-#endif
-	i = 0;
-
-	if (*plenp <= i) sc = 0;
-	else {
-		sc = 1;
-		if (path[i] == _SC_) {
-			for (i++; i < *plenp; i++) if (path[i] != _SC_) break;
-			if (i >= *plenp) sc = 0;
-		}
-	}
-
-	if (isoverwrite) new = realloc2(path, *plenp + sc + flen + 1);
-	else {
-		new = malloc2(*plenp + sc + flen + 1);
-		strncpy(new, path, *plenp);
-	}
-	if (sc) new[(*plenp)++] = _SC_;
-	strncpy2(&(new[*plenp]), file, flen);
-	*plenp += flen;
-	return(new);
+	sp -> s = c_realloc(sp -> s, sp -> len + len, &(sp -> size));
+	memcpy(&(sp -> s[sp -> len]), s, len);
+	sp -> len += len;
+	sp -> s[sp -> len] = '\0';
 }
 
-#ifdef	NODIRLOOP
-static int NEAR _evalwild(argc, argvp, s, fixed, len)
-int argc;
-char ***argvp, *s, *fixed;
-int len;
-#else
-static int NEAR _evalwild(argc, argvp, s, fixed, len, nino, ino)
-int argc;
-char ***argvp, *s, *fixed;
-int len, nino;
-devino_t *ino;
-#endif
+static void NEAR duplwild(dst, src)
+wild_t *dst, *src;
 {
+	memcpy(dst, src, sizeof(wild_t));
+	dst -> fixed.s = (char *)malloc2(src -> fixed.size);
+	memcpy(dst -> fixed.s, src -> fixed.s, src -> fixed.len + 1);
+	dst -> path.s = (char *)malloc2(src -> path.size);
+	memcpy(dst -> path.s, src -> path.s, src -> path.len + 1);
+
 #ifndef	NODIRLOOP
-	devino_t *dupino;
+	if (src -> ino) {
+		dst -> ino = (devino_t *)malloc2(src -> nino
+			* sizeof(devino_t));
+		memcpy(dst -> ino, src -> ino, src -> nino * sizeof(devino_t));
+	}
+#endif	/* !NODIRLOOP */
+}
+
+static void NEAR freewild(wp)
+wild_t *wp;
+{
+	if (wp -> fixed.s) free(wp -> fixed.s);
+	if (wp -> path.s) free(wp -> path.s);
+#ifndef	NODIRLOOP
+	if (wp -> ino) free(wp -> ino);
 #endif
+}
+
+static int NEAR _evalwild(argc, argvp, wp)
+int argc;
+char ***argvp;
+wild_t *wp;
+{
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
 	reg_t *re;
+	wild_t dupl;
+	ALLOC_T flen, plen;
 	char *cp;
-	int i, n, l;
+	int i, n, w, pc, quote, isdir;
 
-	if (!*s) {
-		if (len) {
-			fixed = realloc2(fixed, len + 1 + 1);
-			fixed[len++] = _SC_;
-			fixed[len] = '\0';
-			*argvp = (char **)realloc2(*argvp,
-				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = fixed;
-		}
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
+	if (!*(wp -> s)) {
+		if (!(wp -> fixed.len)) return(argc);
+		*argvp = (char **)realloc2(*argvp,
+			(argc + 2) * sizeof(char *));
+		(*argvp)[argc++] = wp -> fixed.s;
+		wp -> fixed.s = NULL;
 		return(argc);
 	}
 
-#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-	if (!len && _dospath(s)) {
-		fixed = malloc2(2 + 1);
-		fixed[0] = *s;
-		fixed[1] = ':';
-		fixed[2] = '\0';
-# ifdef	NODIRLOOP
-		return(_evalwild(argc, argvp, &(s[2]), fixed, 2));
-# else
-		return(_evalwild(argc, argvp, &(s[2]), fixed, 2, nino, ino));
-# endif
-	}
-#endif
+	flen = wp -> fixed.len;
+	plen = wp -> path.len;
+	quote = wp -> quote;
 
-	n = 0;
-	for (i = 0; s[i]; i++) {
-		if (s[i] == _SC_) break;
-		if (s[i] == '*' || s[i] == '?'
-		|| s[i] == '[' || s[i] == ']') n = 1;
-#ifdef	BSPATHDELIM
-		if (iskanji1(s, i)) i++;
-#endif
-	}
+	if (wp -> fixed.len) addstrbuf(&(wp -> path), _SS_, 1);
 
-	if (!i) {
-		fixed = realloc2(fixed, len + 1 + 1);
-		fixed[len++] = _SC_;
-		fixed[len] = '\0';
-#ifdef	NODIRLOOP
-		return(_evalwild(argc, argvp, &(s[1]), fixed, len));
-#else
-		return(_evalwild(argc, argvp, &(s[1]), fixed, len, nino, ino));
-#endif
-	}
+	for (i = w = 0; wp -> s[i] && wp -> s[i] != _SC_; i++) {
+		pc = parsechar(&(wp -> s[i]), -1,
+			'\0', 0, &(wp -> quote), NULL);
+		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
+			if (!(wp -> flags & EA_STRIPQ))
+				addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			continue;
+		}
+		else if (pc == PC_WORD) {
+			addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
+			i++;
+		}
+		else if (pc == PC_META) {
+			if (wp -> flags & EA_KEEPMETA)
+				addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			if (wp -> s[i + 1] == _SC_) continue;
 
-	if (!n) {
-		fixed = catpath(fixed, s, &len, i, 1);
-		if (Xstat(fixed, &st) < 0) free(fixed);
-		else if (s[i]) {
-			if ((st.st_mode & S_IFMT) != S_IFDIR) free(fixed);
-			else {
-#ifdef	NODIRLOOP
-				return(_evalwild(argc, argvp,
-					&(s[i + 1]), fixed, len));
-#else
-				ino = (devino_t *)realloc2(ino,
-					(nino + 1) * sizeof(devino_t));
-				ino[nino].dev = st.st_dev;
-				ino[nino++].ino = st.st_ino;
-				return(_evalwild(argc, argvp,
-					&(s[i + 1]), fixed, len, nino, ino));
-#endif
+			if (wp -> quote == '\''
+			|| (wp -> quote == '"'
+			&& !strchr(DQ_METACHAR, wp -> s[i + 1]))) {
+				if (!(wp -> flags & EA_KEEPMETA))
+					addstrbuf(&(wp -> fixed),
+						&(wp -> s[i]), 1);
+				addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
 			}
+			i++;
 		}
-		else {
-			*argvp = (char **)realloc2(*argvp,
-				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = fixed;
+		else if (pc == PC_NORMAL && strchr("?*[", wp -> s[i])) {
+			if (w >= 0 && wp -> s[i] == '*') w++;
+			else w = -1;
 		}
+
+		addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+		addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
+	}
+
+	if (!(wp -> s[i])) isdir = 0;
+	else {
+		isdir = 1;
+		addstrbuf(&(wp -> fixed), _SS_, 1);
+	}
+
+	if (!w) {
+		if (wp -> path.len <= plen) w++;
+		else if (stat2(wp -> path.s, &st) < 0) return(argc);
+
+		wp -> s += i;
+		if (isdir) {
+			if (!w && (st.st_mode & S_IFMT) != S_IFDIR)
+				return(argc);
+			(wp -> s)++;
+		}
+
 #ifndef	NODIRLOOP
-		if (ino) free(ino);
+		if (!w) {
+			wp -> ino = (devino_t *)realloc2(wp -> ino,
+				(wp -> nino + 1) * sizeof(devino_t));
+			wp -> ino[wp -> nino].dev = st.st_dev;
+			wp -> ino[(wp -> nino)++].ino = st.st_ino;
+		}
 #endif
-		return(argc);
+		return(_evalwild(argc, argvp, wp));
 	}
 
-	if (i == 2 && s[i] && s[0] == '*' && s[1] == '*') {
-#ifdef	NODIRLOOP
-		argc = _evalwild(argc, argvp, &(s[3]), strdup2(fixed), len);
-#else
-		if (!ino) dupino = NULL;
-		else {
-			dupino = (devino_t *)malloc2(nino * sizeof(devino_t));
-			for (n = 0; n < nino; n++) {
-				dupino[n].dev = ino[n].dev;
-				dupino[n].ino = ino[n].ino;
-			}
-		}
-		argc = _evalwild(argc, argvp,
-			&(s[3]), strdup2(fixed), len, nino, dupino);
-#endif
+	if (w != 2 || !isdir || strcmp(&(wp -> path.s[plen]), "**")) w = -1;
+	wp -> fixed.len = flen;
+	wp -> path.len = plen;
+	wp -> fixed.s[flen] = wp -> path.s[plen] = '\0';
+
+	if (w > 0) {
+		duplwild(&dupl, wp);
+		dupl.s += i + 1;
+		argc = _evalwild(argc, argvp, &dupl);
+		freewild(&dupl);
 		re = NULL;
 	}
-	else if (!(re = regexp_init(s, i))) {
-		if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
-		return(argc);
+	else {
+		cp = malloc2(i + 2);
+		n = 0;
+		if (quote) cp[n++] = quote;
+		memcpy(&(cp[n]), wp -> s, i);
+		n += i;
+		if (wp -> quote) cp[n++] = wp -> quote;
+		re = regexp_init(cp, n);
+		free(cp);
+		if (!re) return(argc);
+		wp -> s += i + 1;
 	}
-	if (!(dirp = Xopendir((len) ? fixed : "."))) {
+
+	if (wp -> path.len) cp = wp -> path.s;
+	else if (wp -> fixed.len) cp = _SS_;
+	else cp = ".";
+
+	if (!(dirp = Xopendir(cp))) {
 		regexp_free(re);
-		if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
 		return(argc);
 	}
+	if (wp -> path.len || wp -> fixed.len)
+		addstrbuf(&(wp -> path), _SS_, 1);
 
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
 
-		l = len;
-		cp = catpath(fixed, dp -> d_name, &l, strlen(dp -> d_name), 0);
-		if (s[i]) {
-			if (Xstat(cp, &st) < 0
+		duplwild(&dupl, wp);
+		n = strlen(dp -> d_name);
+		addstrbuf(&(dupl.fixed), dp -> d_name, n);
+		addstrbuf(&(dupl.path), dp -> d_name, n);
+
+		if (isdir) {
+			if (re) n = regexp_exec(re, dp -> d_name, 1);
+			else n = (*(dp -> d_name) == '.') ? 0 : 1;
+
+			if (!n || Xstat(dupl.path.s, &st) < 0
 			|| (st.st_mode & S_IFMT) != S_IFDIR) {
-				free(cp);
+				freewild(&dupl);
 				continue;
 			}
 
 #ifndef	NODIRLOOP
-			dupino = (devino_t *)malloc2((nino + 1)
-				* sizeof(devino_t));
-			for (n = 0; n < nino; n++) {
-				dupino[n].dev = ino[n].dev;
-				dupino[n].ino = ino[n].ino;
-			}
-			dupino[n].dev = st.st_dev;
-			dupino[n].ino = st.st_ino;
-#endif
 			if (!re) {
-				if (*(dp -> d_name) == '.') {
-					free(cp);
-#ifndef	NODIRLOOP
-					free(dupino);
-#endif
+				for (n = 0; n < dupl.nino; n++)
+					if (dupl.ino[n].dev == st.st_dev
+					&& dupl.ino[n].ino == st.st_ino)
+						break;
+				if (n < dupl.nino) {
+					freewild(&dupl);
 					continue;
 				}
-#ifdef	NODIRLOOP
-				argc = _evalwild(argc, argvp, s, cp, l);
-#else
-				for (n = 0; n < nino; n++)
-					if (ino[n].dev == st.st_dev
-					&& ino[n].ino == st.st_ino) break;
-				if (n < nino) {
-					free(cp);
-					free(dupino);
-				}
-				else argc = _evalwild(argc, argvp,
-						s, cp, l, nino + 1, dupino);
-#endif
 			}
-			else if (!regexp_exec(re, dp -> d_name, 1)) {
-				free(cp);
-#ifndef	NODIRLOOP
-				free(dupino);
-#endif
-			}
-#ifdef	NODIRLOOP
-			else argc = _evalwild(argc, argvp, &(s[i + 1]), cp, l);
-#else
-			else argc = _evalwild(argc, argvp,
-					&(s[i + 1]), cp, l, nino + 1, dupino);
+
+			dupl.ino = (devino_t *)realloc2(dupl.ino,
+				(dupl.nino + 1) * sizeof(devino_t));
+			dupl.ino[dupl.nino].dev = st.st_dev;
+			dupl.ino[(dupl.nino)++].ino = st.st_ino;
 #endif
+
+			addstrbuf(&(dupl.fixed), _SS_, 1);
+			argc = _evalwild(argc, argvp, &dupl);
 		}
-		else if (!re || !regexp_exec(re, dp -> d_name, 1)) free(cp);
-		else {
+		else if (regexp_exec(re, dp -> d_name, 1)) {
 			*argvp = (char **)realloc2(*argvp,
 				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = cp;
+			(*argvp)[argc++] = dupl.fixed.s;
+			dupl.fixed.s = NULL;
 		}
+
+		freewild(&dupl);
 	}
 	Xclosedir(dirp);
 	regexp_free(re);
-	if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-	if (ino) free(ino);
-#endif
+
 	return(argc);
 }
 
@@ -1531,24 +1595,37 @@ CONST VOID_P vp2;
 	return(strpathcmp2(*((char **)vp1), *((char **)vp2)));
 }
 
-char **evalwild(s)
+char **evalwild(s, flags)
 char *s;
+int flags;
 {
+	wild_t w;
 	char **argv;
 	int argc;
 
 	argv = (char **)malloc2(1 * sizeof(char *));
-#ifdef	NODIRLOOP
-	argc = _evalwild(0, &argv, s, NULL, 0);
-#else
-	argc = _evalwild(0, &argv, s, NULL, 0, 0, NULL);
+	w.s = s;
+	w.fixed.s = c_realloc(NULL, 0, &(w.fixed.size));
+	w.path.s = c_realloc(NULL, 0, &(w.path.size));
+	w.fixed.len = w.path.len = (ALLOC_T)0;
+	w.quote = '\0';
+#ifndef	NODIRLOOP
+	w.nino = 0;
+	w.ino = NULL;
 #endif
+	w.flags = flags;
+
+	argc = _evalwild(0, &argv, &w);
+	freewild(&w);
+
 	if (!argc) {
 		free(argv);
 		return(NULL);
 	}
+
 	argv[argc] = NULL;
 	if (argc > 1) qsort(argv, argc, sizeof(char *), cmppath);
+
 	return(argv);
 }
 
@@ -1560,6 +1637,7 @@ char *s;
 	int i;
 
 	for (i = n = 0; s[i]; i++) n += (u_char)(s[i]);
+
 	return(n % MAXHASH);
 }
 
@@ -1585,6 +1663,7 @@ hashlist *next;
 	new -> hits = 0;
 	new -> cost = cost;
 	new -> next = next;
+
 	return(new);
 }
 
@@ -1631,6 +1710,7 @@ hashlist **htable;
 			nextp = &((*nextp) -> next);
 		}
 	}
+
 	return(new);
 }
 
@@ -1642,6 +1722,7 @@ int n;
 
 	for (hp = hashtable[n]; hp; hp = hp -> next)
 		if (!strpathcmp(com, hp -> command)) return(hp);
+
 	return(NULL);
 }
 
@@ -1673,10 +1754,11 @@ int dirok, exe;
 	struct stat st;
 	int d;
 
-	if (Xstat(path, &st) < 0) return(-1);
+	if (stat2(path, &st) < 0) return(-1);
 	d = ((st.st_mode & S_IFMT) == S_IFDIR);
 	if (!exe) return(d);
 	if (d) return(dirok ? d : -1);
+
 	return(Xaccess(path, X_OK));
 }
 
@@ -1701,6 +1783,7 @@ int len, exe;
 		strcpy(&(path[len]), "BAT");
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_BATCH);
 	}
+
 	return(-1);
 }
 #endif	/* MSDOS */
@@ -1713,10 +1796,10 @@ char *com, *search;
 	char *ext;
 #endif
 	char *cp, *tmp, *next, *path;
-	int len, dlen, cost, size, ret, recalc;
+	int len, dlen, cost, size, ret;
 #ifndef	_NOUSEHASH
 	hashlist *hp;
-	int n, duperrno;
+	int n, recalc, duperrno;
 
 	if (!hpp || (!com && !search)) {
 		duperrno = errno;
@@ -1782,7 +1865,9 @@ char *com, *search;
 	}
 #endif	/* CWDINPATH */
 
+#ifndef	_NOUSEHASH
 	recalc = 0;
+#endif
 	if ((next = (search) ? search : getconstvar("PATH"))) {
 		len = strlen(com);
 		size = ret = 0;
@@ -1793,12 +1878,14 @@ char *com, *search;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 			if (_dospath(cp)) next += 2;
 #endif
+#ifndef	_NOUSEHASH
 			if (*next != _SC_) recalc = CM_RECALC;
+#endif
 			next = strchr(next, PATHDELIM);
 			dlen = (next) ? (next++) - cp : strlen(cp);
 			if (!dlen) tmp = NULL;
 			else {
-				tmp = _evalpath(cp, cp + dlen, 1, 1);
+				tmp = _evalpath(cp, cp + dlen, 0);
 				dlen = strlen(tmp);
 			}
 			if (dlen + len + 1 + EXTWIDTH + 1 > size) {
@@ -1860,6 +1947,7 @@ char **argv;
 
 	for (i = 0; i < argc; i++)
 		if (!strpathcmp(argv[i], target)) return(argv[i]);
+
 	return(NULL);
 }
 
@@ -1931,6 +2019,7 @@ char ***argvp;
 	}
 	endpwent();
 #  endif	/* !DEBUG */
+
 	return(argc);
 }
 # endif	/* !NOUID */
@@ -1953,7 +2042,8 @@ int dlen, exe;
 
 	while ((dp = Xreaddir(dirp))) {
 		if ((!len && isdotdir(dp -> d_name))
-		|| strnpathcmp(file, dp -> d_name, len)) continue;
+		|| strnpathcmp(file, dp -> d_name, len))
+			continue;
 		size = strlen(dp -> d_name);
 		if (size + (cp - path) >= MAXPATHLEN) continue;
 		strncpy2(cp, dp -> d_name, size);
@@ -1974,6 +2064,7 @@ int dlen, exe;
 		(*argvp)[argc++] = new;
 	}
 	Xclosedir(dirp);
+
 	return(argc);
 }
 
@@ -1996,11 +2087,12 @@ char ***argvp;
 # endif
 		next = strchr(cp, PATHDELIM);
 		dlen = (next) ? (next++) - cp : strlen(cp);
-		tmp = _evalpath(cp, cp + dlen, 1, 1);
+		tmp = _evalpath(cp, cp + dlen, 0);
 		dlen = strlen(tmp);
 		argc = completefile(file, len, argc, argvp, tmp, dlen, 1);
 		free(tmp);
 	}
+
 	return(argc);
 }
 
@@ -2033,6 +2125,7 @@ int exe;
 # endif
 	else if (exe && dir == path)
 		return(completeexe(path, len, argc, argvp));
+
 	return(completefile(path, len, argc, argvp, ".", 1, exe));
 }
 
@@ -2055,13 +2148,14 @@ char **argv;
 		free(common);
 		return(NULL);
 	}
+
 	return(common);
 }
 #endif	/* !FDSH && !_NOCOMPLETE */
 
-static int NEAR addmeta(s1, s2, stripm, quoted)
+static int NEAR addmeta(s1, s2, quoted, flags)
 char *s1, *s2;
-int stripm, quoted;
+int quoted, flags;
 {
 	int i, j;
 
@@ -2074,7 +2168,7 @@ int stripm, quoted;
 		else if (s2[i] == PMETA) {
 			if (s1) s1[j] = PMETA;
 			j++;
-			if (stripm && s2[i + 1] == PMETA) i++;
+			if (!(flags & EA_KEEPMETA) && s2[i + 1] == PMETA) i++;
 		}
 		else if (!quoted && s2[i] == '\'') {
 			if (s1) s1[j] = PMETA;
@@ -2094,6 +2188,7 @@ int stripm, quoted;
 #endif
 		if (s1) s1[j] = s2[i];
 	}
+
 	return(j);
 }
 
@@ -2114,6 +2209,7 @@ int delim;
 		if (delim) cp[len++] = delim;
 		len = strcpy2(&(cp[len]), argv[i]) - cp;
 	}
+
 	return(cp);
 }
 
@@ -2124,6 +2220,7 @@ char **var;
 
 	if (!var) return(0);
 	for (i = 0; var[i]; i++);
+
 	return(i);
 }
 
@@ -2155,6 +2252,7 @@ int margin;
 	dupl = (char **)malloc2((n + margin + 1) * sizeof(char *));
 	for (i = 0; i < n; i++) dupl[i] = strdup2(var[i]);
 	dupl[i] = NULL;
+
 	return(dupl);
 }
 
@@ -2170,28 +2268,42 @@ int len, spc, flags, *qp, *pqp;
 		}
 		return(PC_CLQUOTE);
 	}
-	if (iskanji1(s, 0)) return(PC_WORD);
+	else if (iskanji1(s, 0)) return(PC_WORD);
 #ifdef	CODEEUC
-	if (isekana(s, 0)) return(PC_WORD);
+	else if (isekana(s, 0)) return(PC_WORD);
 #endif
-	if (*qp == '`') return(PC_BQUOTE);
-	if (*qp == '\'') return(PC_SQUOTE);
-	if (spc && *s == spc) return(*s);
-	if (ismeta(s, 0, *qp, len)) return(PC_META);
-	if ((flags & EA_BACKQ) && *s == '`') {
+#ifdef	NESTINGQUOTE
+	else if (*qp == '`') return(PC_BQUOTE);
+#endif
+	else if (*qp == '\'') return(PC_SQUOTE);
+	else if (spc && *s == spc) return(*s);
+	else if (ismeta(s, 0, *qp, len, flags)) return(PC_META);
+#ifdef	NESTINGQUOTE
+	else if ((flags & EA_BACKQ) && *s == '`') {
 		if (pqp && *qp) *pqp = *qp;
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-	if (*qp) return(PC_DQUOTE);
-	if (!(flags & EA_NOEVALQ) && *s == '\'') {
+#else
+	else if (*qp == '`') return(PC_BQUOTE);
+#endif
+	else if (*qp) return(PC_DQUOTE);
+	else if (!(flags & EA_NOEVALQ) && *s == '\'') {
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-	if (*s == '"') {
+	else if (*s == '"') {
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
+#ifndef	NESTINGQUOTE
+	else if ((flags & EA_BACKQ) && *s == '`') {
+		if (pqp && *qp) *pqp = *qp;
+		*qp = *s;
+		return(PC_OPQUOTE);
+	}
+#endif
+
 	return(PC_NORMAL);
 }
 
@@ -2282,6 +2394,7 @@ int *eolp, *ptrp, qed;
 	if (skipvarvalue(*bufp, ptrp, "}", qed, 1, 0) < 0) return(-1);
 # endif
 #endif	/* BASHSTYLE */
+
 	return(mode);
 }
 
@@ -2299,20 +2412,19 @@ char *next;
 int qed, nonl, nest;
 #endif
 {
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp;
 	int pc, q, len;
 
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	pq = '\0';
 #endif
 	q = '\0';
 	len = strlen(next);
 	while (s[*ptrp]) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
+#ifdef	NESTINGQUOTE
 		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, &pq);
 #else
 		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, NULL);
@@ -2346,6 +2458,7 @@ int qed, nonl, nest;
 
 		(*ptrp)++;
 	}
+
 	return(-1);
 }
 
@@ -2405,6 +2518,7 @@ int plen, mode;
 		free(tmp);
 	}
 	regexp_free(re);
+
 	return(ret);
 }
 
@@ -2423,6 +2537,7 @@ int plen, mode;
 		else new[i] = strdup2(var[i]);
 	}
 	new[i] = NULL;
+
 	return(new);
 }
 #endif	/* !MINIMUMSHELL */
@@ -2472,18 +2587,19 @@ int plen, *modep;
 
 			for (i = j = 0; arglist[i + 1]; i++)
 				j += addmeta(NULL,
-					arglist[i + 1], 0, quoted);
+					arglist[i + 1], quoted, EA_KEEPMETA);
 			if (i <= 0) cp = strdup2("");
 			else {
 				j += (i - 1) * 3;
 				cp = malloc2(j + 1);
-				j = addmeta(cp, arglist[1], 0, quoted);
+				j = addmeta(cp, arglist[1],
+					quoted, EA_KEEPMETA);
 				for (i = 2; arglist[i]; i++) {
 					cp[j++] = quoted;
 					cp[j++] = sp;
 					cp[j++] = quoted;
-					j += addmeta(&(cp[j]),
-						arglist[i], 0, quoted);
+					j += addmeta(&(cp[j]), arglist[i],
+						quoted, EA_KEEPMETA);
 				}
 				cp[j] = '\0';
 			}
@@ -2540,6 +2656,7 @@ int plen, *modep;
 #ifndef	MINIMUMSHELL
 	freevar(new);
 #endif
+
 	return(cp);
 }
 
@@ -2598,6 +2715,7 @@ int s, len, vlen, mode, quoted;
 		if (exitfunc) (*exitfunc)();
 		return(-2);
 	}
+
 	return(mode);
 }
 
@@ -2609,6 +2727,7 @@ int olen, nlen;
 {
 	if (nlen <= olen) return(buf);
 	buf = realloc2(buf, ptr + nlen + (int)strlen(arg) - olen + 1);
+
 	return(buf);
 }
 
@@ -2725,9 +2844,9 @@ int quoted;
 	}
 
 	if (!mode && (c != '@' || !quoted)) {
-		vlen = addmeta(NULL, cp, 0, quoted);
+		vlen = addmeta(NULL, cp, quoted, EA_KEEPMETA);
 		*bufp = insertarg(*bufp, ptr, arg, *argp - arg + 1, vlen);
-		addmeta(&((*bufp)[ptr]), cp, 0, quoted);
+		addmeta(&((*bufp)[ptr]), cp, quoted, EA_KEEPMETA);
 	}
 	else if (!cp) vlen = 0;
 	else if (c == '@' && !*cp && quoted
@@ -2744,6 +2863,7 @@ int quoted;
 
 	ptr += vlen;
 	if (new) free(new);
+
 	return(ptr);
 }
 
@@ -2792,6 +2912,7 @@ char *name;
 	uidlist[maxuid].uid = pwd -> pw_uid;
 	uidlist[maxuid].name = strdup2(pwd -> pw_name);
 	uidlist[maxuid].home = strdup2(pwd -> pw_dir);
+
 	return(&(uidlist[maxuid++]));
 }
 
@@ -2840,6 +2961,7 @@ char *name;
 	gidlist[maxgid].name = strdup2(grp -> gr_name);
 	gidlist[maxgid].gr_mem = duplvar(grp -> gr_mem, -1);
 	gidlist[maxgid].ismem = 0;
+
 	return(&(gidlist[maxgid++]));
 }
 
@@ -2904,10 +3026,10 @@ int rest;
 
 	stripquote(bbuf, 0);
 	if (!(tmp = (*backquotefunc)(bbuf))) return(buf);
-	len = addmeta(NULL, tmp, 1, '\0');
+	len = addmeta(NULL, tmp, '\0', 0);
 	size = *ptrp + len + rest + 1;
 	buf = realloc2(buf, size);
-	addmeta(&(buf[*ptrp]), tmp, 1, '\0');
+	addmeta(&(buf[*ptrp]), tmp, '\0', 0);
 	*ptrp += len;
 	free(tmp);
 
@@ -2944,6 +3066,7 @@ char *gethomedir(VOID_A)
 	if (pwd) return(pwd -> pw_dir);
 # endif	/* !FD */
 #endif	/* !NOUID */
+
 	return(NULL);
 }
 
@@ -2961,7 +3084,7 @@ char **argp;
 	len = ((cp = strdelim(top, 0))) ? (cp - top) : strlen(top);
 	if (!len) cp = gethomedir();
 # ifdef	FD
-	else if (len == sizeof("FD") - 1 && !strnpathcmp(top, "FD", len))
+	else if (len == (int)sizeof("FD") - 1 && !strnpathcmp(top, "FD", len))
 		cp = progpath;
 # endif
 	else {
@@ -3008,6 +3131,7 @@ char **argp;
 	strncpy(&((*bufp)[ptr]), cp, vlen);
 	ptr += vlen;
 	*argp += len;
+
 	return(ptr);
 }
 #endif	/* !MINIMUMSHELL */
@@ -3019,11 +3143,14 @@ int qed, flags;
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char path[MAXPATHLEN], alias[MAXPATHLEN];
 #endif
-#ifdef	BASHSTYLE
+#ifndef	MINIMUMSHELL
+	int prev;
+#endif
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp, *buf, *bbuf;
-	int i, j, pc, prev, q;
+	int i, j, pc, q;
 
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	if (*arg == '"' && (i = strlen(arg)) > 2 && arg[i - 1] == '"') {
@@ -3046,15 +3173,18 @@ int qed, flags;
 	buf = malloc2(i);
 	if (!backquotefunc) flags &= ~EA_BACKQ;
 	bbuf = (flags & EA_BACKQ) ? malloc2(i) : NULL;
-#ifdef	BASHSTYLE
-	pq = '\0';
+#ifndef	MINIMUMSHELL
+	prev =
 #endif
-	prev = q = '\0';
+#ifdef	NESTINGQUOTE
+	pq =
+#endif
+	q = '\0';
 	i = j = 0;
+	cp = arg;
 
-	for (cp = arg; *cp; prev = *(cp++)) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
+	while (*cp) {
+#ifdef	NESTINGQUOTE
 		pc = parsechar(cp, -1, '$', flags, &q, &pq);
 #else
 		pc = parsechar(cp, -1, '$', flags, &q, NULL);
@@ -3120,6 +3250,12 @@ int qed, flags;
 			i = evalhome(&buf, i, &cp);
 #endif
 		else buf[i++] = *cp;
+
+#ifdef	MINIMUMSHELL
+		cp++;
+#else
+		prev = *(cp++);
+#endif
 	}
 #ifndef	BASHSTYLE
 	/* bash does not allow unclosed quote */
@@ -3180,6 +3316,7 @@ char ***argvp, *ifs;
 				(argc-- - n) * sizeof(char *));
 		}
 	}
+
 	return(argc);
 }
 
@@ -3188,61 +3325,31 @@ int argc;
 char ***argvp;
 int flags;
 {
-	char *cp, **wild;
-	ALLOC_T size;
-	int i, j, n, pc, w, quote;
+	char **wild;
+	int i, n;
 
 	for (n = 0; n < argc; n++) {
-		cp = c_realloc(NULL, 0, &size);
-		for (i = j = w = 0, quote = '\0'; (*argvp)[n][i]; i++) {
-			cp = c_realloc(cp, j + 1, &size);
-			pc = parsechar(&((*argvp)[n][i]), -1,
-				'\0', 0, &quote, NULL);
-			if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
-				if (flags & EA_STRIPQ) continue;
-			}
-			else if (pc == PC_WORD) cp[j++] = (*argvp)[n][i++];
-			else if (pc == PC_META) {
-				i++;
-				if (quote
-				&& !strchr(DQ_METACHAR, (*argvp)[n][i]))
-					cp[j++] = PMETA;
-			}
-			else if (pc != PC_NORMAL) /*EMPTY*/;
-			else if (!strchr("?*[", (*argvp)[n][i])) /*EMPTY*/;
-			else if ((wild = evalwild((*argvp)[n]))) {
-				w = countvar(wild);
-				if (w > 1) {
-					*argvp = (char **)realloc2(*argvp,
-						(argc + w) * sizeof(char *));
-					if (!*argvp) {
-						free(cp);
-						freevar(wild);
-						return(argc);
-					}
-					memmove((char *)(&((*argvp)[n + w])),
-						(char *)(&((*argvp)[n + 1])),
-						(argc - n) * sizeof(char *));
-					argc += w - 1;
-				}
-				free((*argvp)[n]);
-				free(cp);
-				memmove((char *)(&((*argvp)[n])),
-					(char *)wild, w * sizeof(char *));
-				free(wild);
-				n += w - 1;
-				break;
-			}
-
-			cp[j++] = (*argvp)[n][i];
+		if (!(wild = evalwild((*argvp)[n], flags))) {
+			stripquote((*argvp)[n], flags);
+			continue;
 		}
 
-		if (!w) {
-			cp[j] = '\0';
-			free((*argvp)[n]);
-			(*argvp)[n] = cp;
+		i = countvar(wild);
+		if (i > 1) {
+			*argvp = (char **)realloc2(*argvp,
+				(argc + i) * sizeof(char *));
+			memmove((char *)(&((*argvp)[n + i])),
+				(char *)(&((*argvp)[n + 1])),
+				(argc - n) * sizeof(char *));
+			argc += i - 1;
 		}
+		free((*argvp)[n]);
+		memmove((char *)(&((*argvp)[n])),
+			(char *)wild, i * sizeof(char *));
+		free(wild);
+		n += i - 1;
 	}
+
 	return(argc);
 }
 
@@ -3262,21 +3369,24 @@ int flags;
 		}
 		else if (pc == PC_WORD) arg[j++] = arg[i++];
 		else if (pc == PC_META) {
-			i++;
 			stripped = 1;
-			if (quote && arg[i] != quote && arg[i] != PMETA)
-				arg[j++] = PMETA;
+			if ((flags & EA_KEEPMETA)
+			|| quote == '\''
+			|| (quote == '"' && !strchr(DQ_METACHAR, arg[i + 1])))
+				arg[j++] = arg[i];
+			i++;
 		}
 
 		arg[j++] = arg[i];
 	}
 	arg[j] = '\0';
+
 	return(stripped);
 }
 
-char *_evalpath(path, eol, uniqdelim, evalq)
+char *_evalpath(path, eol, flags)
 char *path, *eol;
-int uniqdelim, evalq;
+int flags;
 {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char alias[MAXPATHLEN];
@@ -3289,7 +3399,7 @@ int uniqdelim, evalq;
 	else i = strlen(path);
 	cp = strndup2(path, i);
 
-	if (!(tmp = evalarg(cp, '\'', 0))) {
+	if (!(tmp = evalarg(cp, '\'', EA_KEEPMETA))) {
 		*cp = '\0';
 		return(cp);
 	}
@@ -3303,7 +3413,7 @@ int uniqdelim, evalq;
 		pc = parsechar(&(cp[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_CLQUOTE) {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
-			if (!evalq && top >= 0 && ++top < j) {
+			if ((flags & EA_NOEVALQ) && top >= 0 && ++top < j) {
 				tmp[j] = '\0';
 				if (shortname(&(tmp[top]), alias) == alias) {
 					int len;
@@ -3317,39 +3427,44 @@ int uniqdelim, evalq;
 			}
 			top = -1;
 #endif
-			if (evalq) continue;
+			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc == PC_WORD) tmp[j++] = cp[i++];
 		else if (pc == PC_META) {
 			i++;
-			if (!evalq
-			|| (quote && !strchr(DQ_METACHAR, cp[i])))
+			if ((flags & EA_KEEPMETA)
+			|| quote == '\''
+			|| (quote == '"' && !strchr(DQ_METACHAR, cp[i])))
 				tmp[j++] = PMETA;
 		}
 		else if (pc == PC_OPQUOTE) {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 			if (cp[i] == '"') top = j;
 #endif
-			if (evalq) continue;
+			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
-		else if (uniqdelim && cp[i] == _SC_ && c == _SC_) continue;
+		else if (!(flags & EA_NOUNIQDELIM)
+		&& cp[i] == _SC_ && c == _SC_)
+			continue;
 		tmp[j++] = cp[i];
 	}
 	tmp[j] = '\0';
 	free(cp);
+
 	return(tmp);
 }
 
-char *evalpath(path, uniqdelim)
+char *evalpath(path, flags)
 char *path;
-int uniqdelim;
+int flags;
 {
 	char *cp;
 
 	if (!path || !*path) return(path);
 	for (cp = path; *cp == ' ' || *cp == '\t'; cp++);
-	cp = _evalpath(cp, NULL, uniqdelim, 1);
+	cp = _evalpath(cp, NULL, flags);
 	free(path);
+
 	return(cp);
 }
